import os

# For compatibility with Py2.7
try:
    from io import StringIO
except ImportError:
    from StringIO import StringIO

def to_ris_text(entries):
    """
    Convert publication information from WoS XML to RIS format.

    Example
    ==========
    ```python
    ris_recs = wos_parser.rec_info_to_ris(parsed_recs)
    wos_parser.to_ris_text(ris_recs)
    ```

    Parameters
    ==========
    * entries: list, of WoS pubinfo dict entries in RIS format

    Returns
    ==========
    * str, representing publication info in RIS format
    """
    out = StringIO()

    # Markers to indicate WoS sourced RIS file
    out.write("FN Clarivate Analytics Web of Science\n")
    out.write("VR 1.0\n")

    for ent in entries:
        for k, v in ent.items():
            if isinstance(v, list):
                v = [str(i) for i in v if i != ', ' and i is not None]
                v = "\n   ".join(v)
            out.write("{} {}\n".format(k, v))
        # End for
        out.write("ER\n\n")  # End of record marker
    # End for

    return out.getvalue()
# End to_ris_text()

def write_file(text, filename, ext='.txt', overwrite=False):
    """Write string to text file."""
    fn = '{}{}'.format(filename, ext)
    if not os.path.isfile(fn) or overwrite:
        with open(fn, 'w', encoding='utf-8') as outfile:
            outfile.write(text)
            outfile.flush()
    # End if
# End write_txt_file()



RIS_TEMPLATE = {
    'TY': None,  # Publication Type
    'AU': None,  # authors
    'AF': None,  # author_fullnames
    'TI': None,  # pubinfo['item']
    'AB': None,  # Abstract
    'SO': None,  # Source (Journal)
    'LA': None,  # Language
    'DT': None,  # Publication type "{} {}".format(pubinfo['pubtype'], pubinfo['doctype'])
    'DE': None,  # Keywords
    'ID': None,  # Keywords Plus (WoS generated keywords)
    'PY': None,  # Publication Year
    'PD': None,  # Sort Date
    'UT': None,  # WoS UID,
    'DI': None,  # DOI link
    'NC': None,  # Number of citations (custom-defined, not part of RIS specification)
}


def parse_record(rec_data):
    """Parse record data from API request.

    Parameters
    ==========
    * rec_data : dict, record data returned from request
    """
    tmp = RIS_TEMPLATE.copy()

    static_data = rec_data['static_data']
    dynamic_data = rec_data['dynamic_data']

    summary = static_data['summary']
    pub_info = summary['pub_info']

    authors, fullnames = _extract_authors(summary)
    title, pub_src = _extract_manuscript_info(summary)

    full_metadata = static_data['fullrecord_metadata']

    try:
        abstract = full_metadata['abstracts']['abstract']
        abstract_text = abstract['abstract_text']['p']
    except KeyError:
        abstract_text = ""
    # End try

    lang = full_metadata['normalized_languages']['language']
    if not hasattr(lang, 'append'):
        lang = lang['content']
    else:
        for desc in lang:
            if 'primary' in desc['type'].lower():
                lang = desc['content']
                break
        if hasattr(lang, 'append'):
            lang = 'Unknown'
        # End if
    # End if

    pubtype = summary['pub_info']['pubtype']
    doctype = summary['doctypes']['doctype']

    try:
        kws = full_metadata['keywords']['keyword']
    except KeyError:
        kws = []

    try:
        kws_plus = static_data['item']['keywords_plus']['keyword']
    except KeyError:
        kws_plus = []
    try:
    	doi = _extract_doi(dynamic_data['cluster_related']['identifiers']['identifier'])
    except KeyError:
        doi = ""
    nc = dynamic_data['citation_related']['tc_list']['silo_tc']['local_count']

    tmp.update({
        'TY': pub_info['pubtype'],
        'AU': authors,  # authors
        'AF': fullnames,  # author_fullnames
        'TI': title,  # pubinfo['item']
        'AB': abstract_text,  # Abstract
        'SO': pub_src,  # Source (Journal)
        'LA': lang,  # Language
        'DT': "{} {}".format(pubtype, doctype),
        'DE': kws,  # Keywords
        'ID': kws_plus,  # Keywords Plus (WoS generated keywords)
        'PY': pub_info['pubyear'],  # Publication Year
        'PD': pub_info['sortdate'],  # Sort Date
        'UT': rec_data['UID'],  # WoS UID
        'DI': doi,  # DOI or XREF_DOI,
        'NC': nc
    })

    return tmp
# End parse_record()


def extract_ris(data, ris_entries=None):
    if not ris_entries:
        ris_entries = []

    # Parse record set and return
    if 'REC' not in data['Records']['records']:
        print("No records found!")
        return ris_entries

    for rec in data['Records']['records']['REC']:
        ris_entries.append(parse_record(rec))

    return ris_entries


def _extract_authors(pub_summary):
    authors = []
    author_fullname = []
    pub_names = pub_summary['names']['name']
    if isinstance(pub_names, list):
        for person in pub_names:
            if person['role'].lower() == 'author':
                author_fullname.append(person['full_name'])
                authors.append(person['wos_standard'])
    elif isinstance(pub_names, dict):
        # Only 1 entry
        if pub_names['role'].lower() == 'author':
            author_fullname.append(pub_names['full_name'])
            authors.append(pub_names['wos_standard'])
    # End if
    return authors, author_fullname


def _extract_manuscript_info(summary):
    title = None
    pub_src = None
    for it in summary['titles']['title']:
        if it['type'] == 'item':
            title = it['content']

        if it['type'] == 'source':
            pub_src = it['content']
    return title, pub_src


def _extract_doi(identifiers, suitable=None):
    if suitable is None:
        suitable = ['doi', 'xref_doi']

    doi = ""
    if isinstance(identifiers, dict):
        if identifiers['type'] in suitable:
            doi = identifiers['value']
    elif isinstance(identifiers, list):
        for ids in identifiers:
            if ids['type'] in suitable:
                doi = ids['value']
                break
        # End for
    # End if

    return doi



import requests
import json
import yaml
import time
import math
from tqdm.auto import tqdm

from .extractor import extract_ris
from .converter import to_ris_text, write_file


class RESTClient(object):

    BASE_URL = "https://api.clarivate.com/api/wos"
    defaults = {
        'databaseId': 'WOS',
        'lang': 'en',
        'edition': 'WOS+SCI',
        'firstRecord': 1,
        'count': 100,
        'sort': 'PY',
        'optionView': 'FR'
    }

    def __init__(self, config_fn):
        # assert isinstance(config, dict), "Give configuration as a dictionary"

        with open(config_fn, 'r') as fp:
            config = yaml.safe_load(fp)['restful_wos']
            self.config = config

        if 'wos_expanded' in config:
            self.apikey = config['wos_expanded']
        elif 'wos_lite' in config:
            self.apikey = config['wos_lite']
        else:
            raise ValueError("No valid API key could be found in given config file!")
        # End if

        self._req_header = {
            'X-ApiKey': self.apikey,
            'content-type': 'application/json',
        }

    def query(self, query_string, time_span=None, **kwargs):
        defaults = {
            'databaseId': 'WOK',
            'lang': 'en',
            'edition': 'WOS+SCI',
            'firstRecord': 1,
            'count': 100,
            'sort': 'PY',
            'optionView': 'FR'
        }

        search = kwargs if kwargs else {}
        defaults = self.defaults
        for kw in defaults:
            if kw not in kwargs:
                search.update({kw: defaults[kw]})

        search.update({
            'usrQuery': query_string,
        })

        if time_span:
            search.update({'publishTimeSpan': '{}+{}'.format(time_span[0], time_span[1])})

        # Get first 100 records
        resp_data = self.send_query(search)


        result_info = resp_data['QueryResult']
        num_records = result_info['RecordsFound']
        print("Found {} records, retrieving in batches of 100".format(num_records))

        ris_entries = []
        # Strangely, the initial request return has 'Data', but
        # subsequent requests do not.
        ris_entries = extract_ris(resp_data['Data'], ris_entries)

        REQ_MAX = search['count']
        if num_records > REQ_MAX:
            # Need to request more
            query_id = result_info["QueryID"]

            num_requests = int(math.ceil(num_records / REQ_MAX))
            resp_set = []
            with tqdm(total=num_requests, desc='requesting', unit='requests') as pbar:
                pbar.update(1)  # we've already done 1 request
                while search['firstRecord'] + REQ_MAX <= num_records:
                    search.update(
                        {'firstRecord': search['firstRecord'] + REQ_MAX})

                    # resp_data = self.send_query(search, url='{}/query/{}'.format(self.BASE_URL, query_id))
                    resp_set.append(self.send_query(search, url='{}/query/{}'.format(self.BASE_URL, query_id)))
                    pbar.update(1)
                # End while
            # End pbar

            for i, record in tqdm(enumerate(resp_set), desc='processing'):
                try:
                    resp_data = record
                except Exception as e:
                    print(e)
                    raise Exception(e)

                if 'Records' not in resp_data:
                    print("Unexpected return format:")
                    print(resp_data)
                    continue

                if 'records' not in resp_data['Records']:
                    print("No records found for request: {}".format(i))
                    continue
                ris_entries = extract_ris(resp_data, ris_entries)

        return ris_entries
    # End query()

    def send_query(self, search, url=None):
        if not url:
            url = self.BASE_URL
        response = requests.get(url, params=search, headers=self._req_header)
        status = response.status_code
        if status != 200:
            if status == 504:
                # timeout error
                response = self._handle_timeout(search, url)
            else:
                print("RAW:", str(response.headers))
                raise ValueError("Error when sending query.\nStatus Code: {}\nMessage: {}\nParams: {}".format(
                    response.status_code,
                    response.text,
                    search
                ))

        return response.json()
    # End send_query()

    def _handle_timeout(self, search, url):
        time.sleep(5)  # wait 5 seconds...
        return self.send_query(search, url)


if __name__ == '__main__':

    print("Starting!")

    with open('config.yml', 'r') as fp:
        config = yaml.safe_load(fp)['restful_wos']

    client = RESTClient('config.yml')
    resp = client.query('TS=(uncertain* AND (catchment OR watershed OR water))',
                 time_span=('2018-06-01', '2018-12-31'))

    write_file(to_ris_text(resp), 'ris_output', overwrite=True)

    print("Finished!")



#! /usr/bin/env python
from setuptools import setup

if __name__ == "__main__":
    setup(
        name='restful-wos',
        version='1.0.1',
        description='Python package simplifies access to the WoS RESTful API',
        long_description_content_type="text/markdown",
        long_description=open('README.md').read(),
        url='',
        author='Takuya Iwanaga',
        author_email='iwanaga.takuya@anu.edu.au',
        license='ISC',
        packages=['restful_wos'],
        install_requires=[
            'pyyaml',
            'tqdm',
            'pebble',
        ]
    )
